package com.rube;

import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.physics.box2d.*;
import com.badlogic.gdx.physics.box2d.joints.*;
import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.ArrayMap;
import com.badlogic.gdx.utils.Disposable;
import com.zyb.loveball.utils.MathUtil;

import java.util.Iterator;

/**
 * Created by wuhao on 4/13/17.
 */
// load
public class Box2dLoader implements Disposable {
    // track the order of bodies
    Array<Body> bodies = new Array<Body>();
    Array<Joint> joints = new Array<Joint>();
    ArrayMap<Object, CustomProperties> customProperties = new ArrayMap<>();
    Array<ImageData> images = new Array<>();

    Hook hook;

    // 'Hook gives you the chance to change the default load behavior
    public interface Hook {
        void init(WorldData worldData, World world);
        void preLoadBody(BodyData bodyData, BodyDef bodyDef);
        void postLoadBody(BodyData bodyData, Body body);
        void preLoadFixture(FixtureData fixtureData, FixtureDef fixtureDef);
        void postLoadFixture(FixtureData fixtureData, Fixture fixture);
        void finished(WorldData worldData, World world, Box2dLoader loader);
    }
    public static class HookAdapter implements Hook {
        WorldData worldData;
        World world;
        public void init(WorldData worldData, World world){
            this.worldData = worldData;
            this.world = world;
        }

        public void preLoadBody(BodyData bodyData, BodyDef bodyDef) {
        }

        public void postLoadBody(BodyData bodyData, Body body) {
        }

        public void preLoadFixture(FixtureData fixtureData, FixtureDef fixtureDef) {
        }

        public void postLoadFixture(FixtureData fixtureData, Fixture fixture) {
        }

        public void finished(WorldData worldData, World world, Box2dLoader loader) {
        }
    }

    static BodyDef.BodyType int2type(int i) {
        switch (i) {
            case 0:
                return BodyDef.BodyType.StaticBody;
            case 1:
                return BodyDef.BodyType.KinematicBody;
            case 2:
                return BodyDef.BodyType.DynamicBody;
        }
        throw new IllegalArgumentException("unknown type");
    }

    public Hook getHook() {
        return hook;
    }
    public void setHook(Hook hook) {
        this.hook = hook;
    }

    public Array<Body> getBodies() {
        return bodies;
    }
    public Array<Joint> getJoints() {
        return joints;
    }
    public void load(World world, WorldData worldData, Vector2 offset) {
        load(world,worldData,offset,0);
    }

    public void load(World world, WorldData worldData, Vector2 offset, float rotate) {
        bodies.clear();
        joints.clear();
        images.clear();
        customProperties.clear();
        Array<BodyData> body_list = worldData.getBodyList();
        BodyDef bodyDef = new BodyDef();

        if (hook!=null)
            hook.init(worldData, world);

        for (int i=0; i<body_list.size; i++) {
            BodyData bodyData = body_list.get(i);
            bodyDef.type = int2type(bodyData.getType());

            bodyDef.position.set(bodyData.getPosition());
            bodyDef.angle = bodyData.getAngle();
            bodyDef.bullet = bodyData.isBullet();
            bodyDef.fixedRotation = bodyData.isFixedRotation();
            bodyDef.linearDamping = bodyData.getLinearDamping();
            bodyDef.angularDamping = bodyData.getAngularDamping();
            bodyDef.linearVelocity.set(bodyData.getLinearVelocity());
            bodyDef.angularVelocity = bodyData.getAngularVelocity();
            bodyDef.gravityScale = bodyData.getGravityScale();
            if(rotate!=0){
                bodyDef.position.rotate(rotate);
                bodyDef.angle+=(Math.PI*rotate/180F);

            }
            bodyDef.position.add(offset);

            if (hook!=null)
                hook.preLoadBody(bodyData, bodyDef);
            Body body = world.createBody(bodyDef);
            // create fixtures
            createFixtures(body, bodyData, hook);
            if (hook!=null)
                hook.postLoadBody(bodyData, body);
            bodies.add(body);
            if(bodyData.properties!=null){
                customProperties.put(body,bodyData.properties);
            }
        }
        // add joints
        Array<JointData> joint_list = worldData.getJointList();
        for (int i=0; i<joint_list.size; i++) {
            JointData jointData = joint_list.get(i);
            Joint joint = makeJoint(world, bodies, jointData);
            customProperties.put(joint,jointData.properties);
            joints.add(joint);
        }
//        makeImage(worldData.getImageList());
    }


//    private void makeImage(Array<ImageData> dataImages){
//        for(ImageData imageData:dataImages){
//            if(imageData.bodyIndex == -1) continue;
//            Body body = bodies.get(imageData.bodyIndex);
//            ImageData image = imageData.clone();
//            if(body!=null){
//                image.body = body;
//             //   imageData.body = body;
//            }
//
////            ImageUtils utils = new ImageUtils(body,imageData);
////            images.add(utils);
//
//            images.add(image);
//        }
//
//    }


    public static Shape makeShape(ShapeData shapeData) {
        if (shapeData instanceof ShapeData.Circle) {
            ShapeData.Circle data = ((ShapeData.Circle) shapeData);
            CircleShape shape = new CircleShape();
            shape.setRadius(data.radius);
            shape.setPosition(data.center);
            return shape;
        }else if (shapeData instanceof ShapeData.Polygon) {
            ShapeData.Polygon data = ((ShapeData.Polygon) shapeData);
            PolygonShape shape = new PolygonShape();
            float []xx = data.getVerticesX();
            float []yy = data.getVerticesY();
            shape.set(MathUtil.merge(xx, yy));
            return shape;
        }else if (shapeData instanceof ShapeData.Chain) {
            // todo: test
            ShapeData.Chain data = ((ShapeData.Chain) shapeData);
            ChainShape shape = new ChainShape();
            shape.createChain(MathUtil.merge(data.getVerticesX(), data.getVerticesY()));
            if (data.hasNextVertex()) shape.setNextVertex(data.nextVertex);
            if (data.hasPrevVertex()) shape.setNextVertex(data.prevVertex);
            return shape;
        }
        throw new UnsupportedOperationException("no support");
    }


    public void createFixtures(Body body, BodyData bodyData) {
        createFixtures(body, bodyData, null);
    }
    // todo: 1. FixtureDef should be cached
    // todo: 2. filter is not loaded yet
    // add by lilexiang 2017.8.02
    public void createFixtures(Body body, BodyData bodyData, Hook hook) {
        Array<FixtureData> fixtureDatas = bodyData.getFixtures();
        Iterator iterator = fixtureDatas.iterator();
        FixtureDef fixtureDef = new FixtureDef();
        while(iterator.hasNext()){
            FixtureData fixtureData = (FixtureData) iterator.next();
            fixtureDef.density = fixtureData.getDensity();
            fixtureDef.isSensor = fixtureData.isSensor();
            fixtureDef.friction = fixtureData.getFriction();
            fixtureDef.restitution = fixtureData.getRestitution();
            fixtureDef.shape = makeShape(fixtureData.getShapeData());
            if (hook!=null)
                hook.preLoadFixture(fixtureData, fixtureDef);
            Fixture fixture = body.createFixture(fixtureDef);
            if (hook!=null)
                hook.postLoadFixture(fixtureData, fixture);
            Filter filter = new Filter();
            filter.maskBits = fixtureData.filter_maskBits;
            filter.categoryBits = fixtureData.filter_categoryBits;
            filter.groupIndex = fixtureData.filter_groupIndex;
            fixture.setFilterData(filter);
            fixtureDef.shape.dispose();
            if(fixtureData.properties!=null){
                customProperties.put(fixture,fixtureData.properties);
            }
        }

    }


    public static Joint makeJoint(World world, Array<Body> bodies, JointData jointData) {
        String type = jointData.getType();
        if (type.equals("revolute")) {
            RevoluteJointDef jointDef = new RevoluteJointDef();
            jointDef.enableLimit = jointData.is("enableLimit");
            jointDef.enableMotor = jointData.is("enableMotor");
            jointDef.collideConnected = jointData.is("collideConnected");
            jointDef.bodyA = bodies.get(jointData.BodyAindex());
            jointDef.bodyB = bodies.get(jointData.BodyBindex());
            jointDef.localAnchorA.set(jointData.getAnchorA());
            jointDef.localAnchorB.set(jointData.getAnchorB());
            jointDef.referenceAngle = jointDef.bodyB.getAngle() - jointDef.bodyA.getAngle();
            jointDef.lowerAngle = jointData.getFloat("lowerLimit");
            jointDef.upperAngle = jointData.getFloat("upperLimit");
            jointDef.maxMotorTorque = jointData.getFloat("maxMotorTorque");
            jointDef.motorSpeed = jointData.getFloat("motorSpeed");
            return world.createJoint(jointDef);
        }else if (type.equals("distance")) {
            DistanceJointDef jointDef = new DistanceJointDef();
            jointDef.bodyA = bodies.get(jointData.BodyAindex());
            jointDef.bodyB = bodies.get(jointData.BodyBindex());
            jointDef.localAnchorA.set(jointData.getAnchorA());
            jointDef.localAnchorB.set(jointData.getAnchorB());
            jointDef.dampingRatio = jointData.getFloat("dampingRatio", 0);
            jointDef.frequencyHz = jointData.getFloat("frequency", 0);
            jointDef.length = jointData.getFloat("length", 0);
            jointDef.collideConnected = jointData.is("collideConnected");
            return world.createJoint(jointDef);
        }else if(type.equals("wheel")){
            WheelJointDef jointDef = new WheelJointDef();
            jointDef.bodyA = bodies.get(jointData.BodyAindex());
            jointDef.bodyB = bodies.get(jointData.BodyBindex());
            jointDef.localAnchorA.set(jointData.getAnchorA());
            jointDef.localAnchorB.set(jointData.getAnchorB());
            jointDef.collideConnected = jointData.is("collideConnected");
            jointDef.maxMotorTorque = jointData.getFloat("maxMotorTorque");
            jointDef.motorSpeed = jointData.getFloat("motorSpeed");
            jointDef.dampingRatio = jointData.getFloat("dampingRatio");
            jointDef.localAxisA.set(jointData.getLocalAxisA());
            jointDef.frequencyHz = jointData.getFloat("frequency");
            jointDef.enableMotor = jointData.is("enableMotor");
            return world.createJoint(jointDef);
        } else if(type.equals("weld")){
            WeldJointDef jointDef = new WeldJointDef();
            jointDef.bodyA = bodies.get(jointData.BodyAindex());
            jointDef.bodyB = bodies.get(jointData.BodyBindex());
            jointDef.localAnchorA.set(jointData.getAnchorA());
            jointDef.localAnchorB.set(jointData.getAnchorB());
            jointDef.collideConnected = jointData.is("collideConnected");
            jointDef.referenceAngle = jointData.getFloat("refAngle");

            return world.createJoint(jointDef);
        } else if(type.equals("friction")){
            FrictionJointDef jointDef = new FrictionJointDef();
            jointDef.bodyA = bodies.get(jointData.BodyAindex());
            jointDef.bodyB = bodies.get(jointData.BodyBindex());
            jointDef.localAnchorA.set(jointData.getAnchorA());
            jointDef.localAnchorB.set(jointData.getAnchorB());
            jointDef.collideConnected = jointData.is("collideConnected");
            jointDef.maxForce = jointData.getFloat("maxForce");
            jointDef.maxTorque = jointData.getFloat("maxTorque");
            return world.createJoint(jointDef);
        }else if(type.equals("rope")){
            RopeJointDef jointDef = new RopeJointDef();
            jointDef.bodyA = bodies.get(jointData.BodyAindex());
            jointDef.bodyB = bodies.get(jointData.BodyBindex());
            jointDef.localAnchorA.set(jointData.getAnchorA());
            jointDef.localAnchorB.set(jointData.getAnchorB());
            jointDef.collideConnected = jointData.is("collideConnected");
            jointDef.maxLength = jointData.getFloat("maxLength");
            return world.createJoint(jointDef);
        }
        // fixme : I just need revolute joint now. if you need someone else, add it

        throw new RuntimeException("no support");
    }

    public Array<ImageData> getImages() {
        return images;
    }

    public ArrayMap<Object, CustomProperties> getCustomProperties() {
        return customProperties;
    }

    @Override
    public void dispose() {
        bodies.clear();
        joints.clear();
        customProperties.clear();
        images.clear();
        // dispose all disposable objects
    }

}
