package com.rube;

import com.badlogic.gdx.Application;
import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.TextureAtlas;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.utils.ArrayMap;
import com.badlogic.gdx.utils.JsonReader;
import com.badlogic.gdx.utils.JsonValue;
import com.zyb.loveball.utils.ImageComparator;

import java.util.Iterator;

/**
 * Created by Alan on 2017/4/10.
 */
public class JsonLoader {

    final Vector2 v2 = new Vector2();
//    final public static ArrayMap<String, WorldData> datas = new ArrayMap<>();

    Vector2 loadVec2(JsonValue root) {
        v2.x = root.getFloat("x", 0);
        v2.y = root.getFloat("y", 0);
        return v2;
    }

    Vector2 tryLoadVec2(JsonValue root, String key) {
        JsonValue value = root.get(key);
        return value == null ? null : loadVec2(value);
    }

    public WorldData loadWorld(JsonValue root) {
        WorldData worldData = new WorldData();

        Vector2 v = tryLoadVec2(root, "gravity");
        if (v != null)
            worldData.gravity.set(v);
        worldData.allowSleep = root.getBoolean("allowSleep", false);
        worldData.autoClearForces = root.getBoolean("autoClearForces", false);
        worldData.positionIterations = root.getInt("positionIterations", 1);
        worldData.velocityIterations = root.getInt("velocityIterations", 1);

        // load bodies
        for (JsonValue bodyMap = root.getChild("body"); bodyMap != null; bodyMap = bodyMap.next) {
            BodyData bodyData = loadBody(bodyMap);
            worldData.appendBody(bodyData);
        }

        // load images
        for (JsonValue imageMap = root.getChild("image"); imageMap != null; imageMap = imageMap.next()) {
            ImageData imageData = loadImageData(imageMap);
            worldData.appendImage(imageData);
        }

        worldData.getImageList().sort(new ImageComparator());

        // load joints
        for (JsonValue jointsMap = root.getChild("joint"); jointsMap != null; jointsMap = jointsMap.next) {
            JointData jointData = loadJoint(jointsMap);
            worldData.appendJoint(jointData);
        }

        CustomProperties properties = loadProperties(root.get("customProperties"), worldData.toString());
        worldData.setProperties(properties);
        return worldData;
    }

    BodyData loadBody(JsonValue root) {
        BodyData bodyData = new BodyData();
        bodyData.name = root.getString("name", "");
        bodyData.type = root.getInt("type", 0);
        bodyData.gravityScale = root.getFloat("gravityScale", 1f);
        bodyData.awake = root.getBoolean("awake", false);
        bodyData.bullet = root.getBoolean("bullet", false);
        bodyData.massDataI = root.getFloat("massData-I", .1f); //
        bodyData.angle = root.getFloat("angle", 0f);
        bodyData.angularDamping = root.getFloat("angularDamping", 0f);
        bodyData.angularVelocity = root.getFloat("angularVelocity", 0f);
        bodyData.linearDamping = root.getFloat("linearDamping", 0f);
        Vector2 v = tryLoadVec2(root, "linearVelocity");
        if (v != null)
            bodyData.linearVelocity.set(v);
        v = tryLoadVec2(root, "position");
        if (v != null)
            bodyData.position.set(v);

        for (JsonValue fixtureMap = root.getChild("fixture"); fixtureMap != null; fixtureMap = fixtureMap.next) {
            FixtureData fixtureData = loadFixture(fixtureMap);
            bodyData.appendFixture(fixtureData);
        }
        CustomProperties properties = loadProperties(root.get("customProperties"), bodyData.name);
        bodyData.setProperties(properties);
        return bodyData;
    }

    ImageData loadImageData(JsonValue root) {
        ImageData imageData = new ImageData();
        imageData.name = root.getString("name");
        imageData.file = root.getString("file");
        imageData.opacity = root.getFloat("opacity", 1);
        imageData.scale = root.getFloat("scale", 1);
        imageData.angle = root.getFloat("angle", 0);
        imageData.bodyIndex = root.getInt("body");
        imageData.flipX = root.getBoolean("flip", false);
        Vector2 tmp = loadVec2(root.get("center"));
        imageData.centerX = tmp.x;
        imageData.centerY = tmp.y;
        //load color
        JsonValue colorTemp = root.get("colorTint");
        if (colorTemp != null) {
            float[] colorInt = JsonUtil.asFloatArray(colorTemp);
            imageData.color = new Color(colorInt[0] / 255F, colorInt[1] / 255F, colorInt[2] / 255F, colorInt[3] / 255F);
        }
        // load corner x & y
        JsonValue vertices = root.get("corners");
        if (vertices != null) {
            imageData.cornerX = JsonUtil.asFloatArray(vertices.get("x"));
            imageData.cornerY = JsonUtil.asFloatArray(vertices.get("y"));
        }
        CustomProperties properties = loadProperties(root.get("customProperties"), imageData.name);
        if (properties != null)
            imageData.setProperties(properties);

        imageData.cacheFileName();
        return imageData;
    }

    JointData loadJoint(JsonValue root) {
        JointData jointData = new JointData();
        jointData.name = root.getString("name", "");
        jointData.type = root.getString("type", "");
        jointData.bodyA = root.getInt("bodyA");
        jointData.bodyB = root.getInt("bodyB");
        jointData.anchorA.set(loadVec2(root.get("anchorA")));
        jointData.anchorB.set(loadVec2(root.get("anchorB")));

        final String[] bool_keys = {"collideConnected", "enableLimit", "enableMotor",};
        for (String s : bool_keys) {
            if (root.getBoolean(s, false))
                jointData.add(s, true);
        }
        final String[] float_keys = {"jointSpeed", "lowerLimit", "upperLimit", "refAngle",
                "motorSpeed", "maxMotorForce", "maxMotorTorque", "maxForce", "maxTorque",
                "length", "maxLength",
                "dampingRatio", "frequency", "correctionFactor",
        };
        final String[][] float_keys_alias = {
                {"springDampingRatio", "dampingRatio",},
                {"springFrequency", "frequency",},
        };
        for (String s : float_keys) {
            float f = root.getFloat(s, Float.NaN);
            if (!Float.isNaN(f)) {
                jointData.add(s, f);
            }
        }
        for (String[] ss : float_keys_alias) {
            float f = root.getFloat(ss[0], Float.NaN);
            if (!Float.isNaN(f)) {
                jointData.add(ss[1], f);
            }
        }

        Vector2 localAxisA = tryLoadVec2(root, "localAxisA");
        if (localAxisA != null) {
            jointData.setLocalAxisA(localAxisA);
        }

        CustomProperties properties = loadProperties(root.get("customProperties"), jointData.name);
        if (properties != null)
            jointData.setProperties(properties);
        return jointData;
    }

    ShapeData.Circle loadCircle(JsonValue root) {
        ShapeData.Circle circle = new ShapeData.Circle();
        circle.radius = root.getFloat("radius");
        Vector2 v = tryLoadVec2(root, "center");
        if (v != null)
            circle.center.set(v);
        return circle;
    }

    ShapeData.Polygon loadPolygon(JsonValue root) {
        ShapeData.Polygon polygon = new ShapeData.Polygon();
        JsonValue vertices = root.get("vertices");
        if (vertices != null) {
            polygon.vertices_x = JsonUtil.asFloatArray(vertices.get("x"));
            polygon.vertices_y = JsonUtil.asFloatArray(vertices.get("y"));
        }
        return polygon;
    }


    ShapeData.Chain loadChain(JsonValue root) {
        ShapeData.Chain chain = new ShapeData.Chain();
        JsonValue vertices = root.get("vertices");
        if (vertices != null) {
            chain.vertices_x = JsonUtil.asFloatArray(vertices.get("x"));
            chain.vertices_y = JsonUtil.asFloatArray(vertices.get("y"));
        }
        if (root.getBoolean("hasNextVertex", false)) {
            chain.setNextVertex(loadVec2(root.get("nextVertex")));
        }
        if (root.getBoolean("hasPrevVertex", false)) {
            chain.setPrevVertex(loadVec2(root.get("prevVertex")));
        }
        return chain;
    }

    ShapeData loadShape(JsonValue root) {
        JsonValue value;
        value = root.get("circle");
        if (value != null) {
            return loadCircle(value);
        }
        value = root.get("polygon");
        if (value != null) {
            return loadPolygon(value);
        }
        value = root.get("chain");
        if (value != null) {
            return loadChain(value);
        }
        throw new IllegalArgumentException("unknown fixture shape");
    }

    FixtureData loadFixture(JsonValue root) {
        FixtureData fixtureData = new FixtureData();
        fixtureData.name = root.getString("name");
        fixtureData.density = root.getFloat("density", 0f);
        fixtureData.friction = root.getFloat("friction", 0f);
        fixtureData.restitution = root.getFloat("restitution", 0f);
        fixtureData.filter_categoryBits = (short) root.getInt("filter-categoryBits", 1);
        fixtureData.filter_maskBits = (short) root.getInt("filter-maskBits", -1);
        fixtureData.filter_groupIndex = (short) root.getInt("filter-groupIndex", 0);
        fixtureData.sensor = root.getBoolean("sensor", false);
        fixtureData.shapeData = loadShape(root);
        fixtureData.properties = loadProperties(root.get("customProperties"), fixtureData.name);

        return fixtureData;
    }

    CustomProperties loadProperties(JsonValue root, String objectName) {
        CustomProperties prop = new CustomProperties();
        prop.put("Name", objectName);
        if (root != null) {

            for (JsonValue value = root.child(); value != null; value = value.next) {
                String name = value.getString("name");
                if (name.equals("lndex")) {
                    System.out.println("jlkjljl");
                }
                JsonValue s = value.get("string");
                if (s != null) {
                    prop.put(name, s.asString());
                    continue;
                }
                s = value.get("int");
                if (s != null) {
                    prop.put(name, s.asInt());
                    continue;
                }
                s = value.get("float");
                if (s != null) {
                    prop.put(name, s.asFloat());
                    continue;
                }
                s = value.get("bool");
                if (s != null) {
                    prop.put(name, s.asBoolean());
                    continue;
                }
                s = value.get("vec2");
                if (s != null) {
                    prop.put(name, loadVec2(s));
                    continue;
                }
                JsonValue i = value.get("int");
                System.out.println("bla ...");
            }
        }
        return prop;
    }

    public static WorldData load(String file) {
//        WorldData data = datas.get(file);
        WorldData data = null;
        if (data == null || Gdx.app.getType() == Application.ApplicationType.Desktop) {
            JsonValue root = new JsonReader().parse(Gdx.files.internal(file));
            data = new JsonLoader().loadWorld(root);
//            datas.put(file,data);

        }

        return data;
    }

    //
    public static void loadImages(WorldData data, TextureAtlas atlas) {
        Iterator<ImageData> imageIterator = data.imageData.iterator();
        while (imageIterator.hasNext()) {
            ImageData imageData = imageIterator.next();
            imageData.reigon = atlas.findRegion(imageData.getFilename());
        }
    }


//    public static TextureRegion getRegion(String name){
//        TextureRegion region = Asset.instance.vehicle.findRegion(name);
//        if(region == null){
//            region = Asset.instance.vehicle.findRegion(name);
//        }
//        if(region == null){
//            region = Asset.instance.game.findRegion(name);
//        }
//        return region;
//    }


}
