package com.zyb.loveball;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.graphics.OrthographicCamera;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.math.Vector3;
import com.badlogic.gdx.physics.box2d.*;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.Group;
import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.ArrayMap;
import com.badlogic.gdx.utils.ObjectMap;
import com.rube.Box2dLoader;
import com.rube.Drawer;
import com.zyb.loveball.assets.Configuration;
import com.zyb.loveball.assets.LevelData;
import com.zyb.loveball.constants.Constant;
import com.zyb.loveball.objects.*;
import com.zyb.loveball.utils.Log;
import com.zyb.loveball.utils.Matrix4s;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

public class GamePanel extends Group {
    OrthographicCamera camera;
    PaintPanel paintPanel;
    World world;
    Array<Body> bodies = new Array<>();
    Drawer drawer;
    Box2dLoader box2dLoader;
    Scene scene;
    Set<Body> bodyRemoveList = new HashSet<>();
    ArrayMap<Body, Boolean> bodyActiveList = new ArrayMap<>();
    ArrayMap<Body, Vector3> bodyTransformList = new ArrayMap<>();
    ArrayMap<Body, FixtureDef> bodyCreateFixtureList = new ArrayMap<>();
    ArrayMap<Body, Fixture> bodyDestoryFixture = new ArrayMap<>();
    Set<Body> objectRemoveList = new HashSet<>();
    GameState state = GameState.ready;
    private GameState oldState = GameState.ready;
    GameBg bg;
    private Box2DDebugRenderer renderer; // 测试用绘制器
    private UIPanel gameUI;
    PaintController paintController;
    boolean success = false;
    LevelData levelData;

    private Hint hint;

    public GamePanel(float width, float height, UIPanel uiPanel) {
        this.levelData = GameInfo.levelData;
        this.gameUI = uiPanel;
        initCamera(width, height);
        initGame();
        GameTime.reset();
        if (GameInfo.hint || GameInfo.course) {
            hint();
        }
    }

    private void initGame() {
        world = new World(new Vector2(0, -9.8f), true); // 标准重力场
        World.setVelocityThreshold(0);
        drawer = new Drawer();
        box2dLoader = new Box2dLoader();
        box2dLoader.setHook(box2dLoader.getHook());
        bg = new GameBg();
        paintPanel = new PaintPanel(this);
        scene = new Scene(this);
        world.getBodies(bodies);
        world.setContactListener(new WorldContactHandler());
        if (Configuration.box2dDebug) {
            renderer = new Box2DDebugRenderer();
        }
        paintController = new PaintController();
        hint = paintPanel.hint;
    }

    @Override
    protected void setParent(Group parent) {
        super.setParent(parent);
        if (parent == null) return;
        parent.addActor(paintController);
        paintController.setListener(new PaintController.PaintControllerListener() {
            @Override
            public void touchDown(float x, float y) {
                paintPanel.start(x, y);
            }

            @Override
            public void touchDragged(float x, float y) {
                paintPanel.painting(x, y);
            }

            @Override
            public void touchUp(float x, float y) {
                paintPanel.end(x, y);
            }
        });
    }

    private void initCamera(float width, float height) {
        camera = new OrthographicCamera(width, height);
        camera.position.set(camera.viewportWidth / 2, camera.viewportHeight / 2, 0);
        camera.update();
        Matrix4s.game.set(camera.combined);
    }

    public PaintPanel getPaintPanel() {
        return paintPanel;
    }

    public World getWorld() {
        return world;
    }

    public void updateWorldPosition(Vector2 position) {
        gameUI.updateWorldPosition(position);
//        magnifierUtil.update(position);
    }

    public void drawMagnifier(Batch batch) {
        bg.draw(batch, 1);
        scene.draw(batch, 1);
        paintPanel.draw(batch);
    }

    //判断是否重叠，重叠则延长线段到焦点
    public boolean isInBodies(float x, float y) {
        for (Body body : bodies) {
            boolean active = body.isActive();
            if (!active) continue;
            for (Fixture fixture : body.getFixtureList()) {
                if (fixture.testPoint(x, y)) {
                    if (scene.getName(body) == null || !scene.getName(body).equals("fan")) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    public Body createBody(BodyDef bodyDef) {
        Body body = world.createBody(bodyDef);
        bodies.add(body);
        return body;
    }

    public Box2dLoader getBox2dLoader() {
        return box2dLoader;
    }

    private void failure() {
        Log.log("GamePanel", "failure");
        if (state == GameState.failure) return;
        state = GameState.failure;
        gameUI.failure();
    }


    public void end(boolean success, float endDelay) {
        if (state == GameState.end) return;
        Log.log("GamePanel", "end()");
        HelloZombieGame.pauseInputProcessor();
        this.success = success;
        state = GameState.end;
        if (success) {
            int ballId = Configuration.settingData.getBallId();
            if(ballId == Constant.BOMB_ID_1 || ballId == Constant.BOMB_ID_2){
                GameTime.endDelay = 2.3F;
            }else{
                GameTime.endDelay = 1.5F;
            }
            gameUI.screenShot();
        } else {
            GameTime.endDelay = 1.5F;
        }
        if (endDelay != -1) {
            GameTime.endDelay = endDelay;
        }
        gameUI.end();
        paintPanel.end();
    }

    public void end(boolean success) {
        end(success, -1);
    }

    private void success() {
        if (state == GameState.success) return;
        state = GameState.success;
        gameUI.success();
        Log.log("GamePanel", "success");
    }

    public void destroyBody(Body body) {
        if (!world.isLocked()) {
            world.destroyBody(body);
        } else {
            bodyRemoveList.add(body);
        }
    }

    public void destoryFixture(Body body, Fixture fixture) {
        if (!world.isLocked()) {
            body.destroyFixture(fixture);
        } else {
            bodyDestoryFixture.put(body, fixture);
        }
    }

    public void activeBody(Body body, boolean active) {
        if (!world.isLocked()) {
            body.setActive(active);
        } else {
            bodyActiveList.put(body, active);
        }
    }

    public void createFixture(Body body, FixtureDef fixtureDef) {
        if (!world.isLocked()) {
            body.createFixture(fixtureDef);
        } else {
            bodyCreateFixtureList.put(body, fixtureDef);
        }
    }

    public Drawer getDrawer() {
        return drawer;
    }

    public void updateGame(float delta) {
        updateWorld(delta);
        hint.updateHint();
        updateState(delta);
        scene.updateUI(delta);
    }

    float stepTime;
    int velocityIterations = 8; int positionIterations = 3;
    private void updateWorld(float delta) {
        if (state == GameState.gaming || state == GameState.end) {
            preprocess();
            stepTime += delta;
            if(stepTime>2*Constant.TIME_STEP){
                velocityIterations = 4;
                positionIterations = 1;
            }else{
                velocityIterations = 8;
                positionIterations = 3;
            }
            while (stepTime >= Constant.TIME_STEP) {
                stepTime -= Constant.TIME_STEP;
                world.step(Constant.TIME_STEP, velocityIterations, positionIterations);
                scene.updateWorld(delta);
            }
        }
    }

    private void preprocess() {
        Iterator<Body> iterator = bodyRemoveList.iterator();
        while (iterator.hasNext()) {
            Body body = iterator.next();
            bodies.removeValue(body, true);
            body.setContactListener(null);
            world.destroyBody(body);
        }
        Iterator<ObjectMap.Entry<Body, Boolean>> iterator1 = bodyActiveList.iterator();
        while (iterator1.hasNext()) {
            ObjectMap.Entry<Body, Boolean> entry = iterator1.next();
            entry.key.setActive(entry.value);
        }

        Iterator<ObjectMap.Entry<Body, Vector3>> iterator2 = bodyTransformList.iterator();
        while (iterator2.hasNext()) {
            ObjectMap.Entry<Body, Vector3> entry = iterator2.next();
            entry.key.setTransform(entry.value.x, entry.value.y, entry.value.z);
        }

        Iterator<ObjectMap.Entry<Body, FixtureDef>> iterator3 = bodyCreateFixtureList.iterator();
        while (iterator3.hasNext()) {
            ObjectMap.Entry<Body, FixtureDef> entry = iterator3.next();
            entry.key.createFixture(entry.value);
        }

        Iterator<ObjectMap.Entry<Body, Fixture>> iterator4 = bodyDestoryFixture.iterator();
        while (iterator4.hasNext()) {
            ObjectMap.Entry<Body, Fixture> entry = iterator4.next();
            entry.key.destroyFixture(entry.value);
        }
        Iterator<Body> iterator5 = objectRemoveList.iterator();
        while (iterator5.hasNext()) {
            Body body = iterator5.next();
            paintPanel.removeObject(body);
            scene.removeObject(body);
            destroyBody(body);
        }
        bodyActiveList.clear();
        bodyRemoveList.clear();
        bodyTransformList.clear();
        bodyCreateFixtureList.clear();
        bodyDestoryFixture.clear();
        objectRemoveList.clear();
    }


    public void removeObject(Body body) {
        objectRemoveList.add(body);
    }

    private void updateState(float delta) {
        if (!paintPanel.painting) {
            GameTime.unActiveTime += delta;
        }
        if (GameTime.unActiveTime > Constant.handSecond && !GameInfo.hint && (state== GameState.ready || state == GameState.gaming)) {
            gameUI.showHand(true);
        } else {
            if (gameUI != null) {
                gameUI.showHand(false);
            }
        }
        if (!GameInfo.course && GameInfo.failureTimes != 0 && hint.isShowHint() && (state == GameState.gaming || state == GameState.ready)) {
            GameTime.hintTime += delta;
            if (GameTime.hintTime > 3) {
                gameUI.showSkip(true);
                GameTime.hintTime = 0;
            }
        }
        if (state == GameState.end) {
            GameTime.stateTime += delta;
            if (GameTime.stateTime >= GameTime.endDelay) {
                if (success) {
                    success();
                } else {
                    failure();
                }
                GameTime.stateTime = 0;
            }
        }
    }

    public void transformBody(Body body, float x, float y, float rotation) {
        bodyTransformList.put(body, new Vector3(x, y, rotation));
    }

    @Override
    public void draw(Batch batch, float parentAlpha) {
        updateGame(Gdx.graphics.getDeltaTime());
        batch.setProjectionMatrix(camera.combined);
        bg.draw(batch, parentAlpha);
        scene.draw(batch, parentAlpha);
        paintPanel.draw(batch);
        if (paintPanel.painting) {
            if (gameUI != null)
                gameUI.magnifier(true);
        } else {
            if (gameUI != null)
                gameUI.magnifier(false);
        }
        batch.setProjectionMatrix(Matrix4s.ui);
        if (Configuration.box2dDebug) {
            box2dDebug(batch);
        }
    }

    private void box2dDebug(Batch batch) {
        batch.end();
        renderer.render(world, camera.combined);
        batch.begin();
    }

    public void dispose() {
        Log.log("GamePanel","dispose()");
        scene.dispose();
        if (world != null) {
            world.dispose();
            world = null;
        }
        if (paintController != null) {
            paintController.remove();
        }
        bodies.clear();
        scene = null;
        gameUI = null;
        box2dLoader.dispose();
    }

    public void start() {
        if (state == GameState.ready) {
            state = GameState.gaming;
        }
    }

    public Scene getScene() {
        return scene;
    }


    public boolean pause() {
        if (state == GameState.gaming || state == GameState.ready) {
            oldState = state;
            state = GameState.pause;
            return true;
        }
        return false;
    }

    public GameState getState() {
        return state;
    }

    public boolean resume() {
        if (state == GameState.pause) {
            state = oldState;
            return true;
        }
        return false;
    }

    public void updateUI() {
        gameUI.update();
        if (GameInfo.challenge) {
            if (!paintPanel.pen.canWrite()) {
                gameUI.countdown(true, 5);
            }
//            if(GameInfo.chapter ==1){
//                if(paintPanel.pen.getInkLeft() == 0){
//                    gameUI.countdown(true,5);
//                }
//            }else if(GameInfo.chapter ==3){
//                if(!paintPanel.pen.canWrite()){
//
//                }
//            }
        }
    }

    public enum GameState {
        ready, pause, gaming, end, success, failure
    }

    public OrthographicCamera getCamera() {
        return camera;
    }

    public Vector3 getZombieScreenPosition() {
        Vector2 zombiePosition = scene.getZombie().getPosition();
        Vector3 temp = new Vector3();
        temp.set(zombiePosition, 0);
        temp = camera.project(temp);
        return temp;
    }

    public void hint() {
        hint.hint();
        GameInfo.hint = true;
        GameTime.unActiveTime = 0;
        if (gameUI != null)
            gameUI.showHand(false);
    }

    public GameBg getBg() {
        return bg;
    }

    public static class GameTime {
        static float stateTime;
        static float gameTime;
        static float unActiveTime;
        static float endDelay;
        static float hintTime;

        public static void reset() {
            stateTime = 0;
            gameTime = 0;
            unActiveTime = 0;
            endDelay = 0;
            hintTime = 0;
        }
    }

    public void paintEnd() {
        if (GameInfo.challenge) {
            if (!paintPanel.pen.canWrite()) {
                gameUI.countdown(true, 5);
            }
        }
    }

    public float inkLeftPercent() {
        return paintPanel.pen.inkLeftPercent();
    }

    public float getHintLength() {
        return hint.getHintLength();
    }

    public Zombie getZombie() {
        return scene.getZombie();
    }

    public Ball getBall() {
        return scene.getBall();
    }

}
