package com.zyb.loveball.objects;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.physics.box2d.Body;
import com.badlogic.gdx.physics.box2d.Fixture;
import com.badlogic.gdx.physics.box2d.Joint;
import com.badlogic.gdx.scenes.scene2d.Group;
import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.ArrayMap;
import com.badlogic.gdx.utils.DelayedRemovalArray;
import com.rube.CustomProperties;
import com.rube.Drawer;
import com.rube.ImageData;
import com.rube.ImageUtils;
import com.rube.JsonLoader;
import com.rube.WorldData;
import com.zyb.loveball.GamePanel;
import com.zyb.loveball.GameInfo;
import com.zyb.loveball.animations.CommonAnimation;
import com.zyb.loveball.animations.SkeletonAnimation;
import com.zyb.loveball.assets.Assets;
import com.zyb.loveball.assets.Configuration;
import com.zyb.loveball.attributes.DynamicObjectAttribute;
import com.zyb.loveball.constants.Constant;

import java.util.Iterator;

public class Scene {
    GamePanel gamePanel;
    Drawer drawer;
    private DelayedRemovalArray<Joint> joints;
    private DelayedRemovalArray<Body> bodies;
    private ArrayMap<Object, CustomProperties> customProperties;
    private DelayedRemovalArray<ImageUtils> images = new DelayedRemovalArray<>();
    private DelayedRemovalArray<ImageUtils> imagesFrontStatic = new DelayedRemovalArray<>();
    private DelayedRemovalArray<ImageUtils> imagesFrontDynamic = new DelayedRemovalArray<>();
    private Array<SkeletonAnimation> imagesAnimation = new Array();
    StaticScene staticScene;
    Ball ball;
    Zombie zombie;
    Array<BaseObject> frontObjects = new Array<>();
    Array<BaseObject> middleObjects = new Array<>();
    Array<BaseObject> backObjects = new Array<>();

    public Scene(GamePanel panel) {
        this.gamePanel = panel;
        drawer = gamePanel.getDrawer();
        staticScene = new StaticScene(panel);
        load();
    }

    private void load() {
        String sceneFile = "data/chapter/level" + GameInfo.levelId + ".json";
        int index = 0;
        while (!Gdx.files.internal(sceneFile).exists() && index < 100 && GameInfo.levelId > 1) {
            GameInfo.levelId--;
            sceneFile = "data/chapter/level" + GameInfo.levelId + ".json";
            index++;
        }
        WorldData data = JsonLoader.load(sceneFile);
        JsonLoader.loadImages(data, Assets.instance.game);
        gamePanel.getBox2dLoader().load(gamePanel.getWorld(), data, new Vector2(0, 0), 0);
        joints = new DelayedRemovalArray<>();
        bodies = new DelayedRemovalArray<>();
        customProperties = new ArrayMap<>();
        joints.addAll(gamePanel.getBox2dLoader().getJoints());
        bodies.addAll(gamePanel.getBox2dLoader().getBodies());
        customProperties.putAll(gamePanel.getBox2dLoader().getCustomProperties());
        loadImages(data);
        initBodies();
        initImages();
        if (zombie == null || ball == null) return;
        ball.setBombListener(zombie.initBombListener());
        zombie.setZombieListener(ball.initZombieListener());
    }

    protected void initBodies() {
        for (int i = 0; i < bodies.size; i++) {
            Body body = bodies.get(i);
            resetFriction(body);
            String name = getName(body);
            boolean flag = initBoom(name, body);
            if (!flag) {
                flag = initZombie(name, body);
            }
            if (!flag) {
                flag = initStaticObject(name, body);
            }
            if (!flag) {
                initDynamicObject(name, body);
            }

        }
        staticScene.end();
    }

    private void initImages() {
        Iterator<ImageUtils> iterator = images.iterator();
        while (iterator.hasNext()) {
            ImageUtils temp = iterator.next();
            ImageData data = temp.getData();
            if (data.getFilename().equals("spiderWeb")) {
                iterator.remove();
                imagesFrontStatic.add(temp);
            }
            if(data.getFilename().equals("pot")){
                initPotAnimation(data.getCenterX(),data.getCenterY());
                removeImage("pot");
            }
        }
    }
    private void initPotAnimation(float x, float y){
        if (Configuration.poor) return;
        SkeletonAnimation potAnimation = new SkeletonAnimation("animations/pot.skel", Constant.SCALE*0.65f);
        potAnimation.setColor(1, 1, 1, 0.4F);
        imagesAnimation.add(potAnimation);
        potAnimation.setPosition(x, y);
        potAnimation.setyOffset(-0.32f);
        potAnimation.getState().setAnimation(0, "1", true);
    }

    public Zombie getZombie() {
        return zombie;
    }

    private void resetFriction(Body body) {
        Array<Fixture> fixtures = body.getFixtureList();
        for (Fixture temp : fixtures) {
            temp.setFriction(Constant.friction);
        }
        body.resetMassData();
    }

    private boolean initZombie(String name, Body body) {
        if (zombie != null) return false;
        if (name.equals("heart") || name.equals("heart1")) {
            removeImage(body);
            zombie = new Zombie(body, name, gamePanel, 1);
            return true;
        } else if (name.equals("heart2")) {
            removeImage(body);
            zombie = new Zombie(body, name, gamePanel, 2);
            return true;
        } else {
            return false;
        }
    }

    private boolean initBoom(String name, Body body) {
        if (ball != null) return false;
        if (name.equals("ball1") || name.equals("ball")) {
            ball = new Ball(body, gamePanel, 1);
        } else if (name.equals("ball2")) {
            ball = new Ball(body, gamePanel, 2);
        } else if (name.equals("ball3")) {
            ball = new Ball(body, gamePanel, 3);
        }
        if (ball != null) {
            removeImage(body);
            return true;
        }
        return false;
    }

    private boolean initStaticObject(String name, Body body) {
        if (name.contains("_static") || name.endsWith("land")) {
            staticScene.addImage(body);
            removeImage(body);
            return true;
        }
        return false;
    }


    private boolean initDynamicObject(String name, Body body) {
        BaseObject tempObject = null;
        int zIndex = 0;
        if (name.equals("net")) {
            tempObject = new Net(body);
            zIndex = 2;
        } else if (name.equals("fan")) {
            tempObject = new Fan(body, gamePanel, getProperty(body));
        } else if (name.equals("conveyor_belt_body")) {
            tempObject = new ConveyorBelt(body, gamePanel);
        } else if (name.equals("seesaw")) {
            tempObject = new Seesaw(body, gamePanel, customProperties.get(body));
        } else if (name.equals("move_platform_dynamic")) {
            tempObject = new MovePlatform(body, customProperties.get(body), gamePanel);
        } else if (name.startsWith("warpGateIn")) {
            tempObject = createWarpGate(body);
        } else if (name.equals("anti_gravity")) {
            tempObject = new AntiGravity(body, gamePanel);
        } else if (name.equals("ink_bottle")) {
            tempObject = new InkBottle(body, gamePanel);
        } else if (name.equals("border") && body.getFixtureList().size > 1) {
            tempObject = new Border(body, gamePanel);
        } else if (name.equals("vortex")) {
            zIndex = 0;
            tempObject = new Vortex(body, gamePanel);
        } else if (name.equals("area_danger") || (name.equals("border") && body.getFixtureList().size == 1)) {
            tempObject = new AreaDanger(body, gamePanel);
        } else if (name.equals("square_dynamic")) {
            zIndex = 1;
            tempObject = new SquareDynamic(body, gamePanel);
        }else if(name.contains("dynamic")){
            ImageUtils imageUtils = removeImage(body);
            imagesFrontDynamic.add(imageUtils);
        }

        if (name.contains("dynamic")) {
            resetDynamicObjectDenisty(body);
        }

        if (tempObject != null) {
            removeImage(body);
            if (zIndex == 2) {
                frontObjects.add(tempObject);
            } else if(zIndex == 1){
                middleObjects.add(tempObject);
            } else {
                backObjects.add(tempObject);
            }
            return true;
        }
        return false;

    }

    private WarpGate createWarpGate(Body body) {
        WarpGate gate = new WarpGate(body, gamePanel);
        for (int j = 0; j < bodies.size; j++) {
            Body body2 = bodies.get(j);
            String name2 = getName(body2);
            if (name2.equals("warpGateOut")) {
                gate.setOut(body2);
                removeImage(body2);
                break;
            }
        }
        return gate;
    }

    private void resetDynamicObjectDenisty(Body body) {
        Array<Fixture> fixtures = body.getFixtureList();
        for (Fixture temp : fixtures) {
            temp.setDensity(Constant.dynamic_object.density);
        }
        body.resetMassData();
        setDynamicFilter(body);
    }

    public void removeObject(Body body) {
        removeImage(body);
        Iterator<BaseObject> iterator = frontObjects.iterator();
        while (iterator.hasNext()) {
            if (iterator.next().getBody() == body) {
                iterator.remove();
                return;
            }
        }

        iterator = middleObjects.iterator();
        while (iterator.hasNext()) {
            if (iterator.next().getBody() == body) {
                iterator.remove();
                return;
            }
        }

        iterator = backObjects.iterator();
        while (iterator.hasNext()) {
            if (iterator.next().getBody() == body) {
                iterator.remove();
                return;
            }
        }

        Iterator<ImageUtils> iterator1 = imagesFrontDynamic.iterator();
        while(iterator1.hasNext()){
            ImageUtils temp = iterator1.next();
            if(temp.getBody() == body){
                iterator1.remove();
                return;
            }
        }
        if(ball!=null) {
            if (body == ball.getBody()) {
                ball = null;
                gamePanel.end(false);
                return;
            }
        }
        if(zombie!=null) {
            if (body == zombie.getBody()) {
                zombie = null;
                gamePanel.end(false);
                return;
            }
        }
    }

    private void setDynamicFilter(Body body) {
        Array<Fixture> fixtureList = body.getFixtureList();
        for (Fixture fixture : fixtureList) {
            fixture.setFilterData(DynamicObjectAttribute.filter);
        }
    }

    public ImageUtils removeImage(Body body) {
        for (int i = 0; i < images.size; i++) {
            if (body == images.get(i).getBody()) {
                ImageUtils imageUtils = images.get(i);
                images.removeIndex(i);
                return imageUtils;
            }
        }
        return null;
    }
    public ImageUtils removeImage(String name) {
        for (int i = 0; i < images.size; i++) {
            if (images.get(i).getName().equals(name)) {
                ImageUtils imageUtils = images.get(i);
                images.removeIndex(i);
                return imageUtils;
            }
        }
        return null;
    }

    public String getName(Body body) {
        CustomProperties properties = customProperties.get(body);
        if (properties != null) {
            return customProperties.get(body).get("Name");
        } else {
            return null;
        }
    }

    public CustomProperties getProperty(Object object) {
        return customProperties.get(object);
    }

    protected void loadImages(WorldData data) {
        for (int i = 0; i < data.getImageList().size; i++) {
            ImageData temp = data.getImageList().get(i);
            if (temp.getBodyIndex() != -1) {
                images.add(new ImageUtils(bodies.get(temp.getBodyIndex()), temp));
            } else {
                images.add(new ImageUtils(null, temp));
            }
        }
    }

    public void updateWorld(float delta) {
        Iterator<BaseObject> iterator = frontObjects.iterator();
        while (iterator.hasNext()) {
            iterator.next().updateWorld(delta);
        }
        iterator = backObjects.iterator();
        while (iterator.hasNext()) {
            iterator.next().updateWorld(delta);
        }
        if (ball != null)
            ball.updateWorld(delta);
        if (zombie != null)
            zombie.updateWorld(delta);
    }

    public void updateUI(float delta) {
        Iterator<BaseObject> iterator = frontObjects.iterator();
        while (iterator.hasNext()) {
            iterator.next().updateUI(delta);
        }
        iterator = backObjects.iterator();
        while (iterator.hasNext()) {
            iterator.next().updateUI(delta);
        }
        if (ball != null)
            ball.updateUI(delta);
        if (zombie != null)
            zombie.updateUI(delta);
        for(SkeletonAnimation skeletonAnimation : imagesAnimation){
            skeletonAnimation.update();
        }
    }

    public void draw(Batch batch, float parentAlpha) {
        for (ImageUtils imageUtils : images) {
            imageUtils.draw(drawer, batch);
        }
        for(SkeletonAnimation skeletonAnimation : imagesAnimation){
            skeletonAnimation.draw(batch);
        }
        staticScene.draw(batch, parentAlpha);
        for (BaseObject object : backObjects) {
            object.draw(batch, parentAlpha);
        }

        for (BaseObject object : middleObjects) {
            object.draw(batch, parentAlpha);
        }
        for (ImageUtils imageUtils : imagesFrontStatic) {
            imageUtils.draw(drawer, batch);
        }

        for (ImageUtils imageUtils : imagesFrontDynamic) {
            imageUtils.draw(drawer, batch);
        }
        if (zombie != null) {
            zombie.draw(batch, parentAlpha);
        }

        if (ball != null) {
            ball.draw(batch, parentAlpha);
        }

        for (BaseObject object : frontObjects) {
            object.draw(batch, parentAlpha);
        }

    }

    public Ball getBall() {
        return ball;
    }

    public void dispose(){
        joints.clear();
        bodies.clear();
        images.clear();
        customProperties.clear();
        imagesFrontDynamic.clear();
        imagesFrontStatic.clear();
        frontObjects.clear();
        middleObjects.clear();
        backObjects.clear();
        ball = null;
        zombie = null;
    }
}
